/*
 * Copyright (C) 2017 by Andreas Theofilu <andreas@theosys.at>
 * 
 * All rights reserved. No warranty, explicit or implicit, provided.
 *
 * NOTICE:  All information contained herein is, and remains
 * the property of Andreas Theofilu and his suppliers, if any.
 * The intellectual and technical concepts contained
 * herein are proprietary to Andreas Theofilu and its suppliers and
 * may be covered by European and Foreign Patents, patents in process,
 * and are protected by trade secret or copyright law.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Andreas Theofilu.
 */
MODULE_NAME='HomeMaticRPC_Comm'		(dev vdvHM, dev dvClient, dev dvServer, char sIP[], integer nPort)

#include 'HomeMaticRPC'

DEFINE_TYPE
	struct ST_CALL
	{
		char sName[64];			// Name of the procedure to call
		integer nParams;		// Number of parameters the procedure needs
	};

	struct ST_PARAMETERS
	{
		integer nType;			// The type of the parameter
		char sParameter[64];	// The parameter as string content
	};

	struct ST_FIFO
	{
		dev dvDevice;
		char buffer[8192];
	};

DEFINE_CONSTANT
	constant integer MAX_PARAMETERS = 10;
	constant integer MAX_SERVERS = 10;
	constant integer MAX_FIFO = 10;
	constant char HM_VERSION[] = '1.1.2';
	constant integer DEVICE_COMMUNICATING = 151;
	constant integer DATA_INITIALIZED = 252;
	// timelines
	constant long nID_Init = 1;
	constant long nTickInit[] = { 30000, 21000 };		// 21 seconds
	constant long nID_Ping = 2;
	constant long nTickPing[] = { 10000, 10000 };		// 10 seconds
	constant long nID_Fifo = 3;
	constant long nTick_Fifo[] = { 250, 250 };			// 1/4 second

	constant char iperror[20][32]=
	{
		'undefined',						// 1
		'General failure',					// 2
		'undefined',						// 3
		'Unknown Host',						// 4
		'undefined',						// 5
		'connection refused',				// 6
		'connection timed out',				// 7
		'Unknown connection error',			// 8
		'Port already closed',				// 9
		'Binding error',					// 10
		'Listening error',					// 11
		'Socket not connected',				// 12
		'Send to Socket unknown',			// 13
		'Local port already used',			// 14 --> does not close connection!
		'UDP Socket already listening',		// 15
		'too many open sockets (>200)',		// 16
		'Local Port not open'				// 17
	};

	constant char ipservererror[3][40] =
	{
		'invalid server port',
		'invalid value for protocol',
		'unable to open communication port'
	};

DEFINE_VARIABLE
	volatile integer nOnline;
	volatile integer bReInit;							// true: Subscribe to the CCU
	volatile integer nOnlineServer;
	volatile integer nServerStat;
	volatile integer nServerNum;						// Number of real concurrent servers available
	volatile ST_CALL stCall;
	volatile ST_PARAMETERS aParameters[MAX_PARAMETERS];
	volatile ST_PARAMETERS aServerPars[MAX_PARAMETERS];
	volatile long rpcID;
	volatile ST_FIFO _aFifo[MAX_FIFO];
	volatile integer _nFifoPos;
	volatile integer _nFifoPointer;
	volatile integer nFifoMax;
	volatile integer nServerPars;
	volatile integer nDebug;
	volatile char sBufferClient[4096];
	volatile char sBufferServer[65535];

define_function Debug(char mes[])
{
	if (nDebug)
		send_string vdvHM, "'DEBUG:',mes,';'";
}

define_function connect()
{
	if (nOnline)
		return;

	ip_client_open(dvClient.port, sIP, 80, IP_TCP);
	nOnline = 2;
}

define_function connectClient()
{
	if (nOnline)
		return;

	ip_client_open(dvClient.port, sIP, nPort, IP_TCP);
	nOnline = 2;
}

define_function openServer()
{
	if (!nServerStat && nOnlineServer == 0)
	{
		ip_server_open(dvServer.PORT, nPort+1, IP_TCP);
		nOnlineServer = 2;
		nServerStat = 1;
		return;
	}
}

define_function closeAllServer()
{
	if (nServerStat && nOnlineServer == 1)
		ip_server_close(dvServer.PORT);
}

define_function char[64] getDateTime()
{
stack_var char sDate[64];

	switch(day_of_week(LDATE))
	{
		case 1: sDate = 'Sun';
		case 2: sDate = 'Mon';
		case 3: sDate = 'The';
		case 4: sDate = 'Wed';
		case 5: sDate = 'Thu';
		case 6: sDate = 'Fri';
		case 7: sDate = 'Sat';
	}

	sDate = "sDate,', ',format('%02d', date_to_day(LDATE))";

	switch(date_to_month(LDATE))
	{
		case 1:  sDate = "sDate,' Jan'";
		case 2:  sDate = "sDate,' Feb'";
		case 3:  sDate = "sDate,' Mar'";
		case 4:  sDate = "sDate,' Apr'";
		case 5:  sDate = "sDate,' May'";
		case 6:  sDate = "sDate,' Jun'";
		case 7:  sDate = "sDate,' Jul'";
		case 8:  sDate = "sDate,' Aug'";
		case 9:  sDate = "sDate,' Sep'";
		case 10: sDate = "sDate,' Oct'";
		case 11: sDate = "sDate,' Nov'";
		case 12: sDate = "sDate,' Dec'";
	}

	sDate = "sDate,' ',date_to_year(LDATE)";
	sDate = "sDate,format(' %02d:', time_to_hour(TIME)),format('%02d:', time_to_minute(TIME)),format('%02d MET', time_to_second(TIME))";
	return sDate;
}

/*
 * The following functions implement a small fifo buffer who is big enough to
 * hold the RPC calls.
 */
define_function _fifoAdd(dev dvDevice, char mes[])
{
	// Add a new entry
	if (length_string(mes) && dvDevice.PORT > 0)
	{
		_nFifoPos++;

		if (_nFifoPos > MAX_FIFO)
			_nFifoPos = 1;

		_aFifo[_nFifoPos].dvDevice = dvDevice;
		_aFifo[_nFifoPos].buffer = mes;
	}
}

define_function _fifoNext()
{
	if (nOnline == 0)
	{
		connectClient();
		return;
	}

	if (nOnline != 1)
		return;

	if (_nFifoPos == _nFifoPointer)
		return;

	_nFifoPointer++;

	if (_nFifoPointer > MAX_FIFO)
		_nFifoPointer = 1;

	send_string _aFifo[_nFifoPointer].dvDevice,_aFifo[_nFifoPointer].buffer;
	Debug("'--> ',_aFifo[_nFifoPointer].buffer");
	Debug("'--> ',right_string(_aFifo[_nFifoPointer].buffer, 120)");
}

/*
 * The following function is the only answer to a method call from the RPC
 * server.
 * In case of callbacks, which happens after we've subscribed to the server,
 * we are the RPC server for callbacks and have to return an answer.
 */
define_function answerEvent(dev dvDevice)
{
stack_var char head[1024];
stack_var char body[4096];

	head = "'HTTP/1.1 200 OK',$0d,$0a";
	head = "head,'Date: ',getDateTime(),$0d,$0a";
	head = "head,'Server: NetLinx.1.0.0 (VxWorks)',$0d,$0a";
	head = "head,'Connection: close',$0d,$0a";
	head = "head,'Content-Type: text/xml',$0d,$0a";

	body = "'<?xml version="1.0"?>',$0d,$0a";
	body = "body,'<methodResponse><params><param><value></value></param></params></methodResponse>'";

	head = "head,'Content-Length: ',itoa(length_string(body)),$0d,$0a,$0d,$0a";
	send_string dvDevice,"head,body";
}

/*
 * Some basic functions to get or set a value
 */
define_function char[32] getValueType(integer nType)
{
stack_var char sType[32];

	switch(nType)
	{
		case RPC_TYPE_BOOL: 	sType = 'boolean';
		case RPC_TYPE_INT:		sType = 'i4';
		case RPC_TYPE_FLOAT:	sType = 'double';
		case RPC_TYPE_STRING:	sType = 'string';
	}

	return sType;
}

define_function sendServerEvent()
{
	if (nServerPars == 4)
		send_string vdvHM,"aServerPars[3].sParameter,':',aServerPars[2].sParameter,':',itoa(aServerPars[4].nType),':',aServerPars[4].sParameter,';'";
}

define_function integer parseServerEvent(char buf[])
{
stack_var integer i;
stack_var char value[128];
stack_var char sStruct[8192];

	i = 1;

	if (find_string(buf, '<struct>', 1) && find_string(buf, '<name>', 1))
	{
	stack_var integer gpos1;
	stack_var integer gpos2;

		gpos1 = find_string(buf, '<struct>', 1);;

		while (gpos1)
		{
			gpos2 = find_string(buf, '</struct>', gpos1);
			sStruct = mid_string(buf, gpos1 + 8, gpos2 - (gpos1 + 8));

			while (find_string(sStruct, '<value>', 1))
			{
			stack_var integer pos1, pos2;

				pos1 = find_string(sStruct, '<value>', 1) + 7;
				pos2 = find_string(sStruct, '</value>', 1);

				if (pos2 > pos1)
					value = mid_string(sStruct, pos1, pos2 - pos1);
				else
					value = '';

				if (find_string(value, '<string>', 1))
				{
					aServerPars[i].nType = RPC_TYPE_STRING;
					remove_string(value, '>', 1);
					aServerPars[i].sParameter = left_string(value, find_string(value, '</string>', 1) - 1);
				}
				else if (find_string(value, '<boolean>', 1))
				{
					aServerPars[i].nType = RPC_TYPE_BOOL;
					remove_string(value, '>', 1);
					aServerPars[i].sParameter = left_string(value, find_string(value, '</boolean>', 1) - 1);
				}
				else if (find_string(value, '<i4>', 1))
				{
					aServerPars[i].nType = RPC_TYPE_INT;
					remove_string(value, '>', 1);
					aServerPars[i].sParameter = left_string(value, find_string(value, '</i4>', 1) - 1);
				}
				else if (find_string(value, '<double>', 1))
				{
					aServerPars[i].nType = RPC_TYPE_FLOAT;
					remove_string(value, '>', 1);
					aServerPars[i].sParameter = left_string(value, find_string(value, '</double>', 1) - 1);
				}
				else if (find_string(value, '</', 1))
				{
					aServerPars[i].nType = RPC_TYPE_STRING;
					remove_string(value, '>', 1);
					aServerPars[i].sParameter = left_string(value, find_string(value, '</', 1) - 1);
				}
				else
				{
					aServerPars[i].nType = RPC_TYPE_STRING;
					aServerPars[i].sParameter = value;
				}

				remove_string(sStruct, '</value>', 1);
				i++;

				if (i > MAX_PARAMETERS)
					break;
			}

			gpos1 = gpos1 + 8;
			gpos1 = find_string(buf, '<struct>', gpos1);

			if (gpos1)
			{
				nServerPars = i - 1;
				interpretEvent();
				i = 1;
			}
		}
	}

	nServerPars = i - 1;
	return nServerPars;
}

define_function integer findField(char s[])
{
stack_var integer i;

	for (i = 1; i <= nServerPars; i++)
	{
		if (aServerPars[i].sParameter == s)
			return i;
	}

	return 0;
}

define_function integer findFieldType(integer t)
{
stack_var integer i;

	for (i = 1; i <= nServerPars; i++)
	{
		if (aServerPars[i].nType == t)
			return i;
	}

	return 0;
}

define_function interpretEvent()
{
stack_var integer nIdxType;
stack_var integer nIdxName;

	if (!nServerPars)
		return;

	nIdxName = findField('WORKING');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var char addr[32];
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'WORKING:',aServerPars[nIdxName-1].sParameter,':',itoa(aServerPars[nIdxType].nType),':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('STATE');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var char addr[32];
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'EVENT:',aServerPars[nIdxName-1].sParameter,':',itoa(aServerPars[nIdxType].nType),':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('PRESS_SHORT');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'PRESS:SHORT:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('PRESS_LONG');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'PRESS:LONG:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('LEVEL_REAL');
	
	if (!nIdxName)
		nIdxName = findField('LEVEL');

	nIdxType = findFieldType(RPC_TYPE_FLOAT);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'LEVEL:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('TEMPERATURE');
	nIdxType = findFieldType(RPC_TYPE_FLOAT);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'TEMPERATURE:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('HUMIDITY');
	nIdxType = findFieldType(RPC_TYPE_INT);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'HUMIDITY:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('BRIGHTNESS');
	nIdxType = findFieldType(RPC_TYPE_INT);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'BRIGHTNESS:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('MOTION');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'MOTION:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('DIRECTION');
	nIdxType = findFieldType(RPC_TYPE_INT);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'DIRECTION:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('INSTALL_TEST');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'INSTALL_TEST:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('ERROR_OVERLOAD');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'ERROR:OVERLOAD:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('ERROR_OVERHEAT');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'ERROR:OVERHEAT:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('STICKY_UNREACH');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'ERROR:STICKY UNREACH:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('UNREACH');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'ERROR:UNREACH:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('LOWBAT');
	nIdxType = findFieldType(RPC_TYPE_BOOL);

	if (nIdxName > 1 && nIdxType)
	{
	stack_var integer pos1;

		pos1 = find_string(aServerPars[nIdxName-1].sParameter, ':', 1);
		aServerPars[nIdxName-1].sParameter[pos1] = $2e;
		send_string vdvHM,"'ERROR:LOWBAT:',aServerPars[nIdxName-1].sParameter,':',aServerPars[nIdxType].sParameter,';'";

		if (timeline_active(nID_Ping))
			timeline_set(nID_Ping, 0);

		nServerPars = 0;
		return;
	}

	nIdxName = findField('PONG');

	if (nIdxName)
	{
		send_string vdvHM,"'PONG:',aServerPars[nIdxName+1].sParameter,';'";
		nServerPars = 0;
		return;
	}
}

define_function char[4096] createMethodCall()
{
stack_var char method[4096];
stack_var integer i;

	method = "'<methodCall><methodName>',stCall.sName,'</methodName>'";
	method = "method,'<params>'";

	for (i = 1; i <= stCall.nParams; i++)
	{
		if (aParameters[i].nType != RPC_TYPE_NONE)
		{
			method = "method,'<param><value><',getValueType(aParameters[i].nType),'>'";
			method = "method,aParameters[i].sParameter,'</',getValueType(aParameters[i].nType),'></value></param>'";
		}
		else
			method = "method,'<param><value>',aParameters[i].sParameter,'</value></param>'";
	}

	method = "method,'</params>'";
	method = "method,'</methodCall>'";
	return method;
}

define_function sendMessage()
{
stack_var char head[512];
stack_var char body[4096];

	head = "'POST / HTTP/1.1',$0d,$0a";
	head = "head,'Authorization: Basic Og==',$0d,$0a";
	head = "head,'Content-Type: text/xml',$0d,$0a";
	head = "head,'Host: ',sIP,':',itoa(nPort),$0d,$0a";
	body = createMethodCall();
	head = "head,'Content-Length: ',itoa(length_string(body)),$0d,$0a";
	head = "head,'Expect: 100-continue',$0d,$0a";
	head = "head,'Connection: Keep-Alive',$0d,$0a,$0d,$0a";

	connectClient();
	_fifoAdd(dvClient, "head,body");
	Debug("'--> ',body");
}

define_function sendMessageHttp(integer ise_id, char value[])
{
stack_var char head[512];

	if (!ise_id)
		return;

	if (length_string(value))
		head = "'GET /addons/xmlapi/statechange.cgi?ise_id=',itoa(ise_id),'&new_value=',value,' HTTP/1.1',$0d,$0a";
	else
		head = "'GET /addons/xmlapi/state.cgi?datapoint_id=',itoa(ise_id),' HTTP/1.1',$0d,$0a";

	head = "head,'Host:',sIP,$0d,$0a";
	head = "head,'Accept: text/html,*/*',$0d,$0a";
	head = "head,'Accept-Charset: utf-8',$0d,$0a,$0d,$0a";
	connect();
	_fifoAdd(dvClient, head);
}


define_function init()
{
stack_var IP_ADDRESS_STRUCT IPAddress;

	if (!bReInit)
		return;

	get_ip_address(0:1:0, IPAddress);
	rpcID = 20170622;
	stCall.sName = 'init';
	stCall.nParams = 2;
	aParameters[1].nType = RPC_TYPE_STRING;
	aParameters[1].sParameter = "'http://',IPAddress.IPADDRESS,':',itoa(nPort+1),'/'";
	aParameters[2].nType = RPC_TYPE_STRING;
	aParameters[2].sParameter = itoa(rpcID);
	sendMessage();
	bReInit = false;
}

define_function setValue(char addr[32], char value[32], integer type)
{
	stCall.sName = 'setValue';
	stCall.nParams = 3;
	aParameters[1].nType = RPC_TYPE_STRING;
	aParameters[1].sParameter = addr;
	aParameters[2].nType = RPC_TYPE_STRING;

	switch(type)
	{
		case RPC_TYPE_BOOL:		aParameters[2].sParameter = 'STATE';
		case RPC_TYPE_FLOAT:
		case RPC_TYPE_INT:		aParameters[2].sParameter = 'LEVEL';
	}

	aParameters[3].nType = type;
	aParameters[3].sParameter = value;
	sendMessage();
}

define_function getValue(char addr[32], integer type)
{
	stCall.sName = 'getValue';
	stCall.nParams = 2;
	aParameters[1].nType = RPC_TYPE_STRING;
	aParameters[1].sParameter = addr;
	aParameters[2].nType = RPC_TYPE_STRING;

	switch(type)
	{
		case RPC_TYPE_BOOL:		aParameters[2].sParameter = 'STATE';
		case RPC_TYPE_FLOAT:
		case RPC_TYPE_INT:		aParameters[2].sParameter = 'LEVEL';
	}

	sendMessage();
}

define_function ping()
{
	stCall.sName = 'ping';
	stCall.nParams = 1;
	aParameters[1].nType = RPC_TYPE_STRING;
	aParameters[1].sParameter = itoa(rpcID);
	sendMessage();
}

define_function integer parseMethodResponse(char buf[], char value[])
{
stack_var integer type;

	remove_string(buf, '<value>', 1);

	if (find_string(buf, '<string>', 1))
	{
		type = RPC_TYPE_STRING;
		remove_string(buf, '<string>', 1);
		value = left_string(buf, find_string(buf, '</string>', 1) - 1);
		send_string vdvHM,"'EVENT::',itoa(type),':',value,';'";
	}
	else if (find_string(buf, '<boolean>', 1))
	{
		type = RPC_TYPE_BOOL;
		remove_string(buf, '<boolean>', 1);
		value = left_string(buf, find_string(buf, '</boolean>', 1) - 1);
		send_string vdvHM,"'EVENT::',itoa(type),':',value,';'";
	}
	else if (find_string(buf, '<i4>', 1))
	{
		type = RPC_TYPE_INT;
		remove_string(buf, '<i4>', 1);
		value = left_string(buf, find_string(buf, '</i4>', 1) - 1);
		send_string vdvHM,"'EVENT::',itoa(type),':',value,';'";
	}
	else if (find_string(buf, '<int>', 1))
	{
		type = RPC_TYPE_INT;
		remove_string(buf, '<int>', 1);
		value = left_string(buf, find_string(buf, '</int>', 1) - 1);
		send_string vdvHM,"'EVENT::',itoa(type),':',value,';'";
	}
	else if (find_string(buf, '<double>', 1))
	{
		type = RPC_TYPE_FLOAT;
		remove_string(buf, '<double>', 1);
		value = left_string(buf, find_string(buf, '</double>', 1) - 1);
		send_string vdvHM,"'EVENT::',itoa(type),':',value,';'";
	}
	else
	{
		type = RPC_TYPE_STRING;
		remove_string(buf, '>', 1);
		value = left_string(buf, find_string(buf, '</', 1) - 1);
	}

	return type;
}

define_function integer findInstance(dev dvDevice)
{
stack_var integer i;

	for (i = 1; i <= nServerNum; i++)
	{
		if (dvDevice.PORT == dvServer.PORT)
			return i;
	}

	return 0;
}

/*
 * The following functions parse the result of an ordinary call to the
 * "xmlrpc" addon. We use this API because the native XML-RPC API seems to not
 * accept calls who need parameters or at least there's no documentation of how
 * the call have to look like.
 */
define_function parseResult(char str[])
{
stack_var integer ise, idx;
stack_var char value[32];

	remove_string(str, '<changed id="', 1);
	ise = atoi(str);
	remove_string(str, 'new_value="', 1);
	value = left_string(str, find_string(str, '"', 1) - 1);
	send_string vdvHM,"'RESULT:',itoa(ise),':',value,';'";
}

define_function parseState(char str[])
{
stack_var integer ise, idx;
stack_var char value[32];

	remove_string(str, "'ise_id=',$27", 1);
	ise = atoi(str);
	remove_string(str, "'value=',$27", 1);
	value = left_string(str, find_string(str, "$27", 1) - 1);
	send_string vdvHM,"'RESULT:',itoa(ise),':',value,';'";
}

define_function parseAll(char str[])
{
	if (!length_string(str))
		return;

	if (find_string(str, '<stateList>', 1))
		remove_string(str, '</statelist>', 1);
	else if (find_string(str, '<systemVariables>', 1))
		remove_string(str, '</systemVariables>', 1);
	else if (find_string(str, '<result>', 1))
	{
		parseResult(str);
		remove_string(str, '</result>', 1);
	}
	else if (find_string(str, '<state>', 1))
	{
		parseState(str);
		remove_string(str, '</state>', 1);
	}
}

DEFINE_START
	bReInit = true;
	create_buffer dvClient, sBufferClient;
	timeline_create(nID_Fifo, nTick_Fifo, 1, TIMELINE_ABSOLUTE, TIMELINE_REPEAT);

DEFINE_EVENT
	// The Server
	data_event[dvServer]
	{
		online:
		{
			nOnlineServer = 1;
			on[vdvHM,DATA_INITIALIZED];
		}

		offline:
		{
			nOnlineServer = 0;
			nServerStat = 0;
			off[vdvHM,DATA_INITIALIZED];
			openServer();
		}

		onerror:
		{
			if (DATA.NUMBER != 0)
			{
				nOnlineServer = 0;
				nServerStat = 0;
				amx_log(AMX_ERROR,"'HomeMaticRPC_Comm: Error: ',ipservererror[DATA.NUMBER]");
				openServer();
			}
		}

		string:
		{
		stack_var integer pos1, pos2, len, anz, i;

			if (nDebug)
			{
				len = length_string(DATA.TEXT);
				anz = len / 80;

				if ((anz * 80) < len)
					anz++;

				pos1 = 1;

				for (i = 1; i <= anz; i++)
				{
					if (i < anz)
						pos2 = 80;
					else
						pos2 = len - pos1;

					Debug("'S',itoa(i),'<<< ',mid_string(DATA.TEXT, pos1, pos2)");
					pos1 = pos1 + 80;
				}
			}

			if (find_string(DATA.TEXT, '</methodCall>', 1))
			{
				if (length_string(sBufferServer))
					sBufferServer = "sBufferServer,DATA.TEXT";
				else
					sBufferServer = DATA.TEXT;

				if (find_string(sBufferServer, '<methodName>system.multicall</methodName>', 1))
				{
					parseServerEvent(sBufferServer);
					interpretEvent();
					remove_string(DATA.TEXT, '</methodCall>', 1);

					if (timeline_active(nID_Init))
						timeline_set(nID_Init, 0);
				}

				sBufferServer = '';
				answerEvent(DATA.DEVICE);
//				wait 1 ip_server_close(DATA.DEVICE.PORT);
			}
			else
			{
				sBufferServer = "sBufferServer,DATA.TEXT";
				clear_buffer DATA.TEXT;
			}
		}
	}

	// The Client
	data_event[dvClient]
	{
		online:
		{
			nOnline = 1;
			on[vdvHM,DEVICE_COMMUNICATING];

			if (bReInit)
				init();
		}

		offline:
		{
			nOnline = 0;
			off[vdvHM,DEVICE_COMMUNICATING];
		}

		onerror:
		{
			if (DATA.NUMBER != 14)
			{
				if (DATA.NUMBER != 12 && DATA.NUMBER != 13 && DATA.NUMBER != 17)
				{
					off[vdvHM,DEVICE_COMMUNICATING];
					AMX_LOG(AMX_ERROR, "'HomeMaticRPC_Comm: Error: ',iperror[DATA.NUMBER]");
					bReInit = true;
				}

				nOnline = 0;
			}
		}

		string:
		{
		stack_var integer pos1, pos2, len, anz, i;

			if (nDebug)
			{
				len = length_string(DATA.TEXT);
				anz = len / 80;

				if ((anz * 80) < len)
					anz++;

				pos1 = 1;

				for (i = 1; i <= anz; i++)
				{
					if (i < anz)
						pos2 = 80;
					else
						pos2 = len - pos1;

					Debug("'C',itoa(i),'<<< ',mid_string(DATA.TEXT, pos1, pos2)");
					pos1 = pos1 + 80;
				}
			}

			if (find_string(sBufferClient, '</methodResponse>', 1))
			{
			stack_var char value[256];

				parseMethodResponse(sBufferClient, value);
				remove_string(sBufferClient, '</methodResponse>', 1);
				ip_client_close(dvClient.PORT);
			}
			else if ((find_string(sBufferClient, '<stateList>', 1) && find_string(sBufferClient, '</stateList>', 1)) ||
					(find_string(sBufferClient, '<systemVariables>', 1) && find_string(sBufferClient, '</systemVariables>', 1)) ||
					(find_string(sBufferClient, '<result>', 1) && find_string(sBufferClient, '</result>', 1)) ||
					(find_string(sBufferClient, '<state>', 1) && find_string(sBufferClient, '</state>', 1)))
			{
				parseAll(sBufferClient);
				ip_client_close(dvClient.PORT);
			}
		}
	}

	// Virtual device
	data_event[vdvHM]
	{
		online:
		{
			openServer();
			init();

			if (!timeline_active(nID_Init))
				timeline_create(nID_Init, nTickInit, 1, TIMELINE_ABSOLUTE, TIMELINE_REPEAT);

			if (!timeline_active(nID_Ping))
				timeline_create(nID_Ping, nTickPing, 1, TIMELINE_ABSOLUTE, TIMELINE_REPEAT);
		}

		command:
		{
		stack_var char buf[256];

			while (find_string(DATA.TEXT, ';', 1))
			{
				buf = remove_string(DATA.TEXT, ';', 1);
				set_length_string(buf, length_string(buf) - 1);

				select
				{
					// SET VALUE:<addr>:<type>:<value>;
					active (find_string(buf, 'SET VALUE:', 1)):
					{
					stack_var char addr[32];
					stack_var integer type;
					stack_var integer pos;
					stack_var char value[128];

						remove_string(buf, ':', 1);
						addr = left_string(buf, find_string(buf, ':', 1) - 1);
						pos = find_string(addr, '.', 1);

						if (pos)
							addr[pos] = ':';

						remove_string(buf, ':', 1);
						type = atoi(buf);
						remove_string(buf, ':', 1);
						value = buf;
						setValue(addr, value, type);

						if (timeline_active(nID_Ping))
							timeline_set(nID_Ping, 0);
					}

					// GET VALUE:<addr>:<type>;
					active (find_string(buf, 'GET VALUE:', 1)):
					{
					stack_var char addr[32];
					stack_var integer type;
					stack_var integer pos;

						remove_string(buf, ':', 1);
						addr = left_string(buf, find_string(buf, ':', 1) - 1);
						pos = find_string(addr, '.', 1);

						if (pos)
							addr[pos] = ':';

						remove_string(buf, ':', 1);
						type = atoi(buf);
						getValue(addr, type);

						if (timeline_active(nID_Ping))
							timeline_set(nID_Ping, 0);
					}

					active (find_string(buf, 'STATECHANGE:', 1)):
					{
					stack_var integer ise_id;
					stack_var integer type;
					stack_var char value[128];

						remove_string(buf, ':', 1);
						ise_id = atoi(buf);
						remove_string(buf, ':', 1);
						type = atoi(buf);
						remove_string(buf, ':', 1);
						value = buf;
						sendMessageHttp(ise_id, value);

						if (timeline_active(nID_Ping))
							timeline_set(nID_Ping, 0);
					}

					active (find_string(buf, 'STATE:', 1)):
					{
					stack_var integer ise_id;

						remove_string(buf, ':', 1);
						ise_id = atoi(buf);
						sendMessageHttp(ise_id, '');

						if (timeline_active(nID_Ping))
							timeline_set(nID_Ping, 0);
					}

					active (find_string(buf, 'DEBUG:', 1)):
					{
						remove_string(buf, ':', 1);
						nDebug = atoi(buf);
					}
				}
			}
		}
	}

	timeline_event[nID_Init]
	{
		openServer();
//		init();
	}

	timeline_event[nID_Ping]
	{
		ping();
	}

	timeline_event[nID_Fifo]
	{
		_fifoNext();
	}
